==> libpqtypes-1.0.tar.gz: libpqtypes-1.0/src/error.c <==

#include "libpqtypes-int.h"

typedef struct
{
	int size;
	char *msg;
} pqterr_t;

/* When using thread-safe mode, define our static pqterr_t using
 * TLS (Thread-local storage).  There are a couple ways of doing
 * this which is documented below.
 */
#ifdef PQT_THREAD_SAFE
	/* MSVC always uses __declspec(thread) */
#	if defined(PQT_MSVC)
		static __declspec(thread) pqterr_t lasterr = {0};

	/* TLS is defined to __thread by configure (I think GCC 3.3+) */
#	elif defined(HAVE_CONFIG_H) && defined(TLS)
		static TLS pqterr_t lasterr = {0};

	/* Unix-variant, MinGW or Cygwin with no TLS storage class. */
# else
#		define PTHREAD_TLS
#		include <pthread.h>
		/* SCO OpenServer 5 series requires using FSU Threads, which have
		 * slight differences in a couple prototypes.
		 */
#		ifdef __OpenServer__
#			define __initarg any_t val
#		else
#			define __initarg void
#		endif

		static pthread_key_t tlskey_lasterr;
		static pthread_once_t oncekey_lasterr = PTHREAD_ONCE_INIT;

		/* When the thread dies, this will get called. */
		static void tls_free_lasterr(void *value)
		{
			if(value)
			{
				free(value);
				pthread_setspecific(tlskey_lasterr, NULL);
			}
		}

		/* Callback for pthread_once */
		static void tls_init_lasterr(__initarg)
		{
			pthread_key_create(&tlskey_lasterr, tls_free_lasterr);
		}
# endif
#else
	static pqterr_t lasterr = {0};
#endif

static void
vseterror(const char *format, va_list ap, int append);

static pqterr_t *
geterr(void)
{
	pqterr_t *err = NULL;

	/* Thread-safe or not, MSVC and systems with __thread storage class
	 * have a static buffer we can return.  Otherwise, we are in
	 * thread-safe mode and have to use the pthread TLS method.
	 */
#if !defined(PTHREAD_TLS)
	err = &lasterr;

	/* systems requiring pthread TLS keys */
#else
	pthread_once(&oncekey_lasterr, tls_init_lasterr);

	/* SCO OpenServer 5 series uses FSU Threads, which
	 * has minor prototype differences.
	 */
#	ifdef __OpenServer__
		pthread_getspecific(tlskey_lasterr, (any_t *)&err);
#	else
		err = (pqterr_t *)pthread_getspecific(tlskey_lasterr);
#	endif

	if(!err)
	{
		err = (pqterr_t *)malloc(sizeof(pqterr_t));
		memset(err, 0, sizeof(pqterr_t));
		pthread_setspecific(tlskey_lasterr, err);
	}
#endif

	if(err->size == 0)
	{
		err->size = 256;
		err->msg = (char *)malloc(err->size);
		*err->msg = 0;
	}

	return err;
}

const char *
PQgeterror(void)
{
	pqterr_t *err = geterr();
	return (const char *)err->msg;
}

void
PQseterror(const char *format, ...)
{
	/* clear error message by passing in NULL, PQseterror(NULL).*/
	if(!format)
	{
		pqterr_t *err = geterr();
		*err->msg = 0;
	}
	else
	{
		va_list ap;
		va_start(ap, format);
		vseterror(format, ap, FALSE);
	}
}

/* errorf() callback for PGtypeArgs, see PQputf() and PQgetf().
 * Always returns -1.
 */
int
pqt_argserrorf(PGtypeArgs *args, const char *format, ...)
{
	va_list ap;
	char fqtn[200];

	if(!args || !format || !*format)
		return -1;

	pqt_fqtn(fqtn, sizeof(fqtn),
		args->typhandler->typschema,
		args->typhandler->typname);

	/* prepend header */
	PQseterror("%s[pos:%d] - ", fqtn, args->typpos);

	/* message from type handler */
	va_start(ap, format);
	vseterror(format, ap, TRUE);
	return -1;
}

static void
vseterror(const char *format, va_list ap, int append)
{
	int n;
	int curlen=0;
	char *p;
	va_list ap2;
	pqterr_t *err = geterr();

	if(append)
		curlen = (int)strlen(err->msg);
	else
		*err->msg = 0;

  while(1)
	{
		va_copy(ap2, ap);
		n = pqt_vsnprintf(err->msg+curlen, err->size-curlen, format, ap2);
		va_end(ap2);

		/* success */
		if(n > -1)
			break;

		/* need more space */
		n = err->size * 2;
		if(!(p = pqt_realloc(err->msg, n)))
		{
			*err->msg = 0;
			break;
		}

		err->msg = p;
		err->size = n;
	}
}

==> libpqtypes-1.2b.tar.gz: libpqtypes-1.2b/src/error.c <==

/*
 * error.c
 *   The functions in this file represent the libpqtypes error
 *   system.  It offers the API user the ability to set/get errors.
 *   The error system uses a per-thread global error, implemented
 *   using TLS (via declspec(thread), __thread or pthread TLS).  For
 *   non-PQT_THREAD_SAFE builds, a static buffer is used.
 *
 * Copyright (c) 2008 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

typedef struct
{
	int size;
	char *msg;
} pqterr_t;

/* When using thread-safe mode, define our static pqterr_t using
 * TLS (Thread-local storage).  There are a couple ways of doing
 * this which is documented below.
 */
#ifdef PQT_THREAD_SAFE
	/* MSVC always uses __declspec(thread) */
#	if defined(PQT_MSVC)
		static __declspec(thread) pqterr_t lasterr = {0};

	/* TLS is defined to __thread by configure (I think GCC 3.3+) */
#	elif defined(HAVE_CONFIG_H) && defined(TLS)
		static TLS pqterr_t lasterr = {0};

	/* Unix-variant, MinGW or Cygwin with no TLS storage class. */
# else
#		define PTHREAD_TLS /* using pthread's TLS */
#		include <pthread.h>

		/* System is using FSU Threads, like SCO OpenServer 5, which
		 * has minor prototype differences.
		 */
#		ifdef PTHREAD_FSU
#			define __initarg void *val
#		else
#			define __initarg void
#		endif

		static pthread_key_t tlskey_lasterr;
		static pthread_once_t oncekey_lasterr = PTHREAD_ONCE_INIT;

		/* When the thread dies, this will get called. */
		static void tls_free_lasterr(void *value)
		{
			if(value)
			{
				pqterr_t *e = (pqterr_t *)value;
				if(e->msg)
					free(e->msg);
				free(value);
				pthread_setspecific(tlskey_lasterr, NULL);
			}
		}

		/* Callback for pthread_once */
		static void tls_init_lasterr(__initarg)
		{
			pthread_key_create(&tlskey_lasterr, tls_free_lasterr);
		}
# endif
#else
	static pqterr_t lasterr = {0};
#endif

static void
vseterror(const char *format, va_list ap, int append);

static pqterr_t *
geterr(void)
{
	pqterr_t *err = NULL;

	/* Thread-safe or not, MSVC and systems with __thread storage class
	 * have a static buffer we can return.  Otherwise, we are in
	 * thread-safe mode and have to use the pthread TLS method.
	 */
#if !defined(PTHREAD_TLS)
	err = &lasterr;

	/* systems requiring pthread TLS keys */
#else
	pthread_once(&oncekey_lasterr, tls_init_lasterr);

	/* System is using FSU Threads, like SCO OpenServer 5, which
	 * has minor prototype differences.
	 */
#	ifdef PTHREAD_FSU
		pthread_getspecific(tlskey_lasterr, (void **)&err);
#	else
		err = (pqterr_t *)pthread_getspecific(tlskey_lasterr);
#	endif

	if(!err)
	{
		err = (pqterr_t *)malloc(sizeof(pqterr_t));
		memset(err, 0, sizeof(pqterr_t));
		pthread_setspecific(tlskey_lasterr, err);
	}
#endif

	if(err->size == 0)
	{
		err->size = 256;
		err->msg = (char *)malloc(err->size);
		*err->msg = 0;
	}

	return err;
}

const char *
PQgeterror(void)
{
	pqterr_t *err = geterr();
	return (const char *)err->msg;
}

void
PQseterror(const char *format, ...)
{
	/* clear error message by passing in NULL, PQseterror(NULL).*/
	if(!format)
	{
		pqterr_t *err = geterr();
		*err->msg = 0;
	}
	else
	{
		va_list ap;
		va_start(ap, format);
		vseterror(format, ap, FALSE);
	}
}

/* errorf() callback for PGtypeArgs, see PQputf() and PQgetf().
 * Always returns -1.
 */
int
pqt_argserrorf(PGtypeArgs *args, const char *format, ...)
{
	va_list ap;
	char fqtn[200];

	if(!args || !format || !*format)
		return -1;

	pqt_fqtn(fqtn, sizeof(fqtn), args->typhandler->typschema,
		args->typhandler->typname);

	/* put the header */
	PQseterror("%s[pos:%d] - ", fqtn, args->typpos);

	/* append message from type handler */
	va_start(ap, format);
	vseterror(format, ap, TRUE);
	return -1;
}

static void
vseterror(const char *format, va_list ap, int append)
{
	int n;
	int curlen=0;
	char *p;
	va_list ap2;
	pqterr_t *err = geterr();

	if(append)
		curlen = (int)strlen(err->msg);
	else
		*err->msg = 0;

  while(1)
	{
		va_copy(ap2, ap);
		n = pqt_vsnprintf(err->msg+curlen, err->size-curlen, format, ap2);
		va_end(ap2);

		/* success */
		if(n > -1)
			break;

		/* need more space */
		n = err->size * 2;
		if(!(p = pqt_realloc(err->msg, n)))
		{
			*err->msg = 0;
			break;
		}

		err->msg = p;
		err->size = n;
	}
}
